#!/usr/bin/env python3
"""Hub functions."""

import sys
from pathlib import Path

from tolkein import tofile
from tolkein import tolog

from .attributes import fetch_types
from .hub import process_names_file

LOGGER = tolog.logger(__name__)


def validate_types_file(types_file, dir_path, es, types_name, opts):
    """Validate types file."""
    LOGGER.info("Validating YAML file %s", types_file)
    try:
        types = tofile.load_yaml(str(types_file.resolve()))
    except Exception:
        LOGGER.error("Unable to open types file %s", str(types_file.resolve()))
        sys.exit(1)
    if "attributes" in types:
        try:
            attributes = fetch_types(es, types_name, opts)
        except Exception:
            attributes = {}
        for key, entry in types["attributes"].items():
            if key in attributes:
                entry = {**attributes[key], **entry}
            enum = entry.get("constraint", {}).get("enum", [])
            if enum:
                entry["constraint"]["enum"] = [value.lower() for value in enum]

    names = None
    data = None
    if "file" in types:
        if "name" in types["file"]:
            if "taxonomy" not in types:
                LOGGER.error("Types file contains no taxonomy information")
                sys.exit(1)
            datafile = Path(dir_path) / types["file"]["name"]
            data = tofile.open_file_handle(datafile)
            if data is None:
                LOGGER.error("Data file '%s' could not de opened for reading", datafile)
                sys.exit(1)
        defaults = {"attributes": {}, "metadata": {}}
        for key, value in types["file"].items():
            if key.startswith("display") or key.startswith("taxon"):
                defaults["attributes"].update({key: value})
            elif key.startswith("source"):
                defaults["attributes"].update({key: value})
                defaults["metadata"].update({key: value})
        types.update({"defaults": defaults})
        if "name" in types["file"]:
            names = process_names_file(
                types, Path(dir_path) / "names" / types["file"]["name"]
            )
    return types, data, names



#!/usr/bin/env python3
"""BlobToolKit functions."""

import ujson
from tolkein import tofetch
from tolkein import tolog

LOGGER = tolog.logger(__name__)

BTK_API = "https://blobtoolkit.genomehubs.org/api/v1"
BTK_VIEW = "https://blobtoolkit.genomehubs.org/view"


def fetch_btk_datasets(params):
    """Fetch BlobToolKit taxon entries."""
    root = params.get("root", "Eukaryota")
    url = "%s/search/%s" % (BTK_API, root)
    page = tofetch.fetch_url(url)
    datasets = ujson.decode(page)
    return datasets


def stream_btk_datasets(root=None):
    """Stream BlobToolKit taxon entries."""
    params = {"root": root}
    datasets = fetch_btk_datasets(params)
    if datasets:
        for dataset in datasets:
            yield dataset["id"], dataset


def extract_btk_stats(meta):
    """Extract BlobToolKit stats to top level."""
    summaryStats = meta.pop("summaryStats", {})
    meta["source"] = "BlobToolKit"
    meta["sourceSlug"] = meta["id"]
    meta["sourceStub"] = "https://blobtoolkit.genomehubs.org/view/dataset/"
    # meta["xref"] = "BTK:%s" % meta["id"]
    if "busco" in summaryStats:
        for busco_lineage, busco_stats in summaryStats["busco"].items():
            meta["busco_lineage"] = busco_lineage
            meta["busco_string"] = busco_stats["string"]
            meta["busco_complete"] = busco_stats["c"] / busco_stats["t"] * 100
            break
    if "stats" in summaryStats:
        meta["nohit"] = summaryStats["stats"]["noHit"] * 100
        try:
            meta["target"] = summaryStats["stats"]["target"] * 100
        except KeyError:
            pass
    if "baseComposition" in summaryStats:
        meta["at_percent"] = summaryStats["baseComposition"]["at"] * 100
        meta["gc_percent"] = summaryStats["baseComposition"]["gc"] * 100
        meta["n_percent"] = summaryStats["baseComposition"]["n"] * 100
    try:
        if len(meta["taxon_name"]) > len(meta["species"]):
            meta["subspecies"] = meta["taxon_name"]
    except KeyError:
        meta["species"] = meta["taxon_name"]


def describe_btk_files(meta):
    """Generate analysis descriptions and links for BlobToolKit."""
    plots = ["cumulative", "snail"]
    summaryStats = meta.get("summaryStats", {})
    if "readMapping" in summaryStats and summaryStats["readMapping"]:
        plots.append("blob")
    files = []
    for plot in plots:
        if plot == "blob":
            url = "%s/image/%s/%s/circle?format=png" % (BTK_API, meta["id"], plot)
        else:
            url = "%s/image/%s/%s?format=png" % (BTK_API, meta["id"], plot)
        obj = {
            "name": "%s.png" % plot,
            "url": url,
            "link": "%s/%s/dataset/%s/%s" % (BTK_VIEW, meta["id"], meta["id"], plot),
            "analysis_id": "btk-%s" % meta["id"],
            "description": "a %s plot from BlobToolKit analysis %s"
            % (plot, meta["id"]),
            "title": "%s plot %s" % (plot, meta["id"]),
            "command": "blobtoolkit pipeline",
            "assembly_id": meta["accession"],
            "taxon_id": str(meta["taxid"]),
            "analysis": {
                "name": "BlobToolKit",
                "title": "BlobToolKit analysis of %s" % meta["accession"],
                "description": "Analysis of public assembly %s using BlobToolKit"
                % meta["accession"],
            },
        }
        files.append(obj)
    return files


def btk_parser(_params, opts, *args, **kwargs):
    """Parse BlobToolKit assemblies."""
    parsed = []
    analyses = []
    for root in opts["btk-root"]:
        for key, meta in stream_btk_datasets(root):
            files = describe_btk_files(meta)
            analyses += files
            extract_btk_stats(meta)
            # TODO: set xref
            # TODO: add to file/analysis index
            # print(lineage)
            # print(
            #     "%s: %s (%s)" % (key, meta["name"]["value"], meta["rankLabel"]["value"])
            # )
            # identifiers = fetch_gbif_identifiers(key, xrefs=opts["gbif-xref"])
            # parsed += prepare_xref_rows(key, meta, lineage)
            parsed.append(meta)
    return (parsed, analyses)



#!/usr/bin/env python3

"""Assembly indexing methods."""

import copy
import sys
from collections import defaultdict

from tolkein import tolog

from .es_functions import document_by_id
from .es_functions import index_stream
from .es_functions import query_keyword_value_template
from .es_functions import query_value_template
from .hub import add_attribute_values
from .hub import chunks
from .hub import index_templator
from .taxon import add_taxonomy_info_to_meta
from .taxon import find_or_create_taxa
from .taxon import index_template as taxon_index_template
from .taxon import stream_taxa
from .taxonomy import index_template as taxonomy_index_template

LOGGER = tolog.logger(__name__)


def index_template(taxonomy_name, opts):
    """Index template (includes name, mapping and types)."""
    parts = ["assembly", taxonomy_name, opts["hub-name"], opts["hub-version"]]
    template = index_templator(parts, opts)
    return template


def get_list_entries_by_dict_value(values, list_of_dicts, *, key="key"):
    """Get entries from a list of dicts by key."""
    entries = []
    for entry in list_of_dicts:
        if key in entry and entry[key] in values:
            entries.append(entry)
    return entries


def get_list_indices_by_dict_value(values, list_of_dicts, *, key="key"):
    """Get indices from a list of dicts by key."""
    indices = []
    for idx, entry in enumerate(list_of_dicts):
        if key in entry and entry[key] in values:
            indices.append(idx)
    return indices


def add_identifiers_to_list(existing, new, *, blanks=set({"NA"})):
    """Add identifiers to a list if they do not already exist."""
    # TODO: include source?
    identifiers = defaultdict(dict)
    for entry in existing:
        identifiers[entry["class"]][entry["identifier"]] = True
    for entry in new:
        id_class = entry["class"]
        identifier = entry["identifier"]
        if (
            identifier not in blanks
            and id_class not in identifiers
            and identifier not in identifiers[id_class]
        ):
            existing.append({"identifier": identifier, "class": id_class})
            identifiers[id_class][identifier] = True


def lookup_assemblies_by_assembly_id(es, values, template, *, return_type="list"):
    """Retrieve existing taxa from index."""
    assemblies = []
    if return_type == "dict":
        assemblies = {}
    res = document_by_id(es, values, template["index_name"])
    for key, value in res.items():
        key = key.replace("assembly-", "")
        if return_type == "list":
            assemblies.append(key)
        else:
            assemblies.update({key: value})
    return assemblies


def add_identifiers_and_attributes_to_assemblies(
    es, data, opts, *, template, taxon_template, blanks=set(["NA"])
):
    """Add identifiers and attributes to assemblies."""
    all_assemblies = {}
    taxon_id_by_asm = {}
    for taxon_id, assemblies in data.items():
        for asm in assemblies:
            asm["taxon_id"] = taxon_id
            for identifier in asm["identifiers"]:
                if identifier["class"] == "assembly_id":
                    all_assemblies[identifier["identifier"]] = asm
                    taxon_id_by_asm[identifier["identifier"]] = taxon_id
    for values in chunks(list(all_assemblies.keys()), 500):
        assembly_res = query_keyword_value_template(
            es,
            "attributes_identifiers_by_keyword_value",
            "assembly_id",
            values,
            index=template["index_name"],
        )
        if assembly_res is None:
            assembly_res = {"responses": [{} for value in values]}
        taxon_ids = set()
        for value in values:
            taxon_ids.add(taxon_id_by_asm[value])
        taxa = find_or_create_taxa(
            es,
            opts,
            taxon_ids=taxon_ids,
            taxon_template=taxon_template,
            asm_by_taxon_id=data,
        )
        for index, response in enumerate(assembly_res["responses"]):
            doc = {
                "_id": "assembly-%s" % values[index],
                "_source": {
                    "assembly_id": values[index],
                    "identifiers": [],
                    "attributes": [],
                },
            }
            if "hits" in response and response["hits"]["total"]["value"] == 1:
                doc = response["hits"]["hits"][0]
            assembly_data = all_assemblies[doc["_source"]["assembly_id"]]
            identifiers = copy.deepcopy(doc["_source"]["identifiers"])
            attributes = copy.deepcopy(doc["_source"]["attributes"])
            if "attributes" in assembly_data:
                attributes = attributes + assembly_data["attributes"]
            if "identifiers" in assembly_data:
                identifiers = identifiers + assembly_data["identifiers"]
            add_identifiers_to_list(
                doc["_source"]["identifiers"], identifiers, blanks=blanks
            )
            if "attributes" not in doc["_source"] or not doc["_source"]["attributes"]:
                doc["_source"]["attributes"] = []
            add_attribute_values(doc["_source"]["attributes"], attributes, raw=False)
            doc["_source"]["taxon_id"] = assembly_data["taxon_id"]
            try:
                add_taxonomy_info_to_meta(
                    doc["_source"], taxa[assembly_data["taxon_id"]]["_source"]
                )
                yield doc["_id"], doc["_source"]
            except KeyError:
                LOGGER.warning(
                    "Taxon ID %s was not found in the taxonomy"
                    % assembly_data["taxon_id"]
                )


def add_assembly_attributes_to_taxon(
    batch, asm_by_taxon_id, taxa, *, index_name, shared_attributes
):
    """Add assembly attributes to taxon."""
    for taxon_id, assemblies in asm_by_taxon_id.items():
        if taxon_id in taxa:
            taxon_attributes = defaultdict(list)
            if "attributes" not in taxa[taxon_id]:
                taxa[taxon_id]["attributes"] = []
            for idx in assemblies:
                asm_meta = batch[idx]
                attributes = get_list_entries_by_dict_value(
                    shared_attributes, asm_meta["attributes"]
                )
                for attr in attributes:
                    taxon_attr = {**attr}
                    del taxon_attr["key"]
                    taxon_attr["source_index"] = index_name
                    taxon_attr["source_doc_id"] = (
                        "assembly-%s" % asm_meta["assembly_id"]
                    )
                    taxon_attributes[attr["key"]].append(taxon_attr)
            for key, values in taxon_attributes.items():
                indices = get_list_indices_by_dict_value(
                    {key}, taxa[taxon_id]["attributes"]
                )
                if len(indices) == 1:
                    idx = indices[0]
                    # TODO: test if values are already present
                    taxa[taxon_id]["attributes"][idx]["values"] += values
                else:
                    taxa[taxon_id]["attributes"].append({"key": key, "values": values})


def collate_unique_key_value_indices(key, list_of_dicts):
    """Collate indices of unique key values in a list of dicts."""
    unique_key_values = set()
    entry_indices_by_key_value = defaultdict(list)
    for idx, entry in enumerate(list_of_dicts):
        if entry[key] not in unique_key_values:
            unique_key_values.add(entry[key])
        entry_indices_by_key_value[entry[key]].append(idx)
    return list(unique_key_values), entry_indices_by_key_value


def get_taxa_to_create(
    es,
    batch,
    opts,
    *,
    taxonomy_name="ncbi",
    taxon_ids=None,
    asm_by_taxon_id=None,
    template=None,
    shared_attributes=None
):
    """Create a dict of taxa to create."""
    taxa_to_create = {}
    if not taxon_ids:
        return {}
    taxonomy_template = taxonomy_index_template(taxonomy_name, opts)
    taxonomy_res = query_value_template(
        es,
        "taxonomy_node_by_taxon_id",
        taxon_ids,
        taxonomy_template["index_name"],
    )
    if taxonomy_res is None:
        LOGGER.error(
            "Could not connect to taxonomy index '%s'",
            taxonomy_template["index_name"],
        )
        sys.exit(1)
    ancestors = set()
    for taxonomy_result in taxonomy_res["responses"]:
        if taxonomy_result["hits"]["total"]["value"] == 1:
            source = taxonomy_result["hits"]["hits"][0]["_source"]
            taxa_to_create[source["taxon_id"]] = source
            for ancestor in source["lineage"]:
                ancestors.add(ancestor["taxon_id"])
            for idx in asm_by_taxon_id[source["taxon_id"]]:
                add_taxonomy_info_to_meta(batch[idx], source)
    add_assembly_attributes_to_taxon(
        batch,
        asm_by_taxon_id,
        taxa_to_create,
        index_name=template["index_name"],
        shared_attributes=shared_attributes,
    )
    taxonomy_res = query_value_template(
        es,
        "taxonomy_node_by_taxon_id",
        list(ancestors),
        taxonomy_template["index_name"],
    )
    if taxonomy_res and "responses" in taxonomy_res:
        for taxonomy_result in taxonomy_res["responses"]:
            if taxonomy_result["hits"]["total"]["value"] == 1:
                source = taxonomy_result["hits"]["hits"][0]["_source"]
                taxa_to_create[source["taxon_id"]] = source
    return taxa_to_create


def preprocess_batch(es, batch, opts, *, taxonomy_name="ncbi"):
    """Preprocess a batch of assembly metadata to add/update taxonomy information."""
    template = index_template(taxonomy_name, opts)
    taxon_template = taxon_index_template(taxonomy_name, opts)
    # TODO: find shared attributes programatically
    shared_attributes = {}
    taxon_ids, asm_by_taxon_id = collate_unique_key_value_indices("taxon_id", batch)
    taxon_res = query_keyword_value_template(
        es,
        "attributes_by_keyword_value",
        "taxon_id",
        taxon_ids,
        taxon_template["index_name"],
    )
    taxa = {}
    if taxon_res is not None:
        for taxon_result in taxon_res["responses"]:
            if (
                "error" not in taxon_result
                and taxon_result["hits"]["total"]["value"] == 1
            ):
                source = taxon_result["hits"]["hits"][0]["_source"]
                taxa[source["taxon_id"]] = source
                for idx in asm_by_taxon_id[source["taxon_id"]]:
                    add_taxonomy_info_to_meta(batch[idx], source)
    taxa_to_update = {}
    taxon_ids = []
    for taxon_id in asm_by_taxon_id.keys():
        if taxon_id in taxa:
            taxa_to_update[taxon_id] = {}
            if "attributes" in taxa[taxon_id]:
                taxa_to_update[taxon_id]["attributes"] = taxa[taxon_id]["attributes"]
        else:
            taxon_ids.append(taxon_id)
    add_assembly_attributes_to_taxon(
        batch,
        asm_by_taxon_id,
        taxa_to_update,
        index_name=template["index_name"],
        shared_attributes=shared_attributes,
    )
    taxa_to_create = get_taxa_to_create(
        es,
        batch,
        opts,
        taxonomy_name=taxonomy_name,
        taxon_ids=taxon_ids,
        asm_by_taxon_id=asm_by_taxon_id,
        template=template,
        shared_attributes=shared_attributes,
    )
    to_create = len(taxa_to_create.keys())
    to_update = len(taxa_to_update.keys())
    LOGGER.info(
        "%d taxa to create, %d to update",
        to_create,
        to_update,
    )
    if to_create > 0:
        index_stream(es, taxon_template["index_name"], stream_taxa(taxa_to_create))
    if to_update > 0:
        index_stream(
            es,
            taxon_template["index_name"],
            stream_taxa(taxa_to_update),
            _op_type="update",
        )
    return to_create, to_update


def stream_assemblies(assemblies):
    """Stream list of assemblies for indexing."""
    for asm in assemblies:
        yield "assembly-%s" % asm["assembly_id"], asm


def set_top_level_meta(raw_meta):
    """Set top level assembly metadata."""
    top_level = {
        "assembly_id": raw_meta["assembly_id"],
        "taxon_id": raw_meta["taxon_id"],
    }
    return top_level


# def index(es, opts, *, metadata_name="insdc", taxonomy_name="ncbi"):
#     """Index all assemblies descended from root."""
#     parsers = {"insdc": parse_insdc_metadata}
#     sources = {"insdc": "INSDC"}
#     parser = parsers.get(metadata_name.lower(), None)
#     if parser is None:
#         LOGGER.warning("No parser available for %s metadata", metadata_name)
#         return None
#     LOGGER.info("Indexing %s metadata", metadata_name)
#     template = index_template(taxonomy_name, opts)

#     stream = parser(opts)
#     batch = []
#     iteration = 1
#     while True:
#         stop = False
#         try:
#             raw_meta = next(stream)
#             attributes = add_attributes(
#                 raw_meta,
#                 template["types"]["attributes"],
#                 attributes=[],
#                 source=sources.get(metadata_name, metadata_name),
#             )
#             identifiers = add_attributes(
#                 raw_meta,
#                 template["types"]["identifiers"],
#                 attributes=[],
#                 source=sources.get(metadata_name, metadata_name),
#                 attr_type="identifiers",
#             )
#             top_level = set_top_level_meta(raw_meta)
#             asm_meta = {
#                 **top_level,
#                 "attributes": attributes,
#                 "identifiers": identifiers,
#             }
#             batch.append(asm_meta)
#         except StopIteration:
#             stop = True
#         if len(batch) == opts["es-batch"] or stop:
#             # TODO: set attributes
#             LOGGER.info("Processing batch %d with %d assemblies", iteration, len(batch))

#             created, updated = preprocess_batch(
#                 es, batch, opts, taxonomy_name=taxonomy_name
#             )
#             LOGGER.info("Indexing %d assemblies", len(batch))
#             assembly_stream = stream_assemblies(batch)
#             index_stream(es, template["index_name"], assembly_stream)
#             batch = []
#             iteration += 1
#         if stop:
#             break
#     return



#!/usr/bin/env python3

"""File indexing methods."""

import hashlib
import os
import urllib
from datetime import date
from mimetypes import guess_type
from pathlib import Path
from shutil import copyfile

import filetype
from PIL import Image
from tolkein import tofetch
from tolkein import tofile
from tolkein import tolog

from .analysis import index_template as analysis_index_template
from .assembly import index_template as assembly_index_template
from .assembly import lookup_assemblies_by_assembly_id
from .es_functions import document_by_id
from .es_functions import index_stream
from .hub import index_templator
from .taxon import find_or_create_taxa
from .taxon import index_template as taxon_index_template
from .taxon import list_ancestors

LOGGER = tolog.logger(__name__)

TODAY = date.today().strftime("%Y-%m-%d")


def index_template(taxonomy_name, opts):
    """Index template (includes name, mapping and types)."""
    parts = ["file", taxonomy_name, opts["hub-name"], opts["hub-version"]]
    template = index_templator(parts, opts)
    return template


def force_copy(sourcename, destname, symlink=True):
    """Create or replace a symlink to a file."""
    localfile = Path(destname)
    os.makedirs(Path(destname).parent, exist_ok=True)
    try:
        localfile.unlink()
    except FileNotFoundError:
        pass
    if symlink:
        localfile.symlink_to(Path(sourcename))
    else:
        copyfile(Path(sourcename), localfile)


def process_image_file(infile, filename, opts, *, dest_dir="./", attrs=None):
    """Process an image file for indexing."""
    if attrs is None:
        attrs = {}
    filepath = Path(filename)
    dimensions = (100, 100)
    thumbname = "%s.thm%s" % (filepath.stem, filepath.suffix)
    with Image.open(infile) as im:
        attrs.update(
            {
                "preview_name": thumbname,
                "preview_mime_type": attrs["mime_type"],
                "format": im.format,
                "size_pixels": "%sx%s" % (im.width, im.height),
            }
        )
        if filename != thumbname:
            try:
                im.thumbnail(dimensions)
                os.makedirs("%s/%s" % (opts["hub-path"], dest_dir), exist_ok=True)
                im.save("%s/%s/%s" % (opts["hub-path"], dest_dir, thumbname))
            except OSError as err:
                print(err)
                LOGGER.warn("Cannot create thumbnail for '%s'", infile)
        attrs.update()
    return attrs


def set_file_meta_defaults(opts):
    """Set default values for file metadata."""
    defaults = {}
    for key in {
        "taxon-id",
        "assembly-id",
        "analysis-id",
        "file-title",
        "file-description",
    }:
        value = opts.get(key, False)
        if value:
            defaults.update({key.replace("file-", "").replace("-", "_"): value})
    return defaults


def make_local_copy(infile, filepath, filename, attrs, opts, local):
    """Link or copy file to local directory."""
    try:
        dest_dir = str(filepath.parent.relative_to(opts["hub-path"]))
    except ValueError:
        dest_dir = "files"
        try:
            dest_dir += "/taxon-" + attrs.get("taxon_id", "NA")
        except TypeError:
            dest_dir += "/taxon-MULTI"
        try:
            dest_dir += "/assembly-" + attrs.get("assembly_id", "NA")
        except TypeError:
            dest_dir += "/assembly-MULTI"
        dest_dir += "/analysis-" + attrs.get("analysis_id", "all")
        localname = "%s/%s/%s" % (opts["hub-path"], dest_dir, filename)
        if local == "symlink":
            force_copy(infile, localname)
        elif local == "copy":
            force_copy(infile, localname, False)
    return dest_dir


def process_file(
    infile,
    opts,
    *,
    file_template=None,
    analysis_template=None,
    filename=None,
    meta=None,
    local="symlink"
):
    """Process a file for indexing."""
    filepath = Path(infile)
    if filename is None:
        filename = filepath.name
    if meta is None:
        meta = {}
    attrs = {
        "name": filename,
        "size_bytes": os.path.getsize(infile),
        **set_file_meta_defaults(opts),
        **meta,
    }
    dest_dir = make_local_copy(infile, filepath, filename, attrs, opts, local)
    attrs.update({"location": dest_dir})
    kind = filetype.guess(infile)
    if kind is not None:
        attrs.update({"extension": kind.extension, "mime_type": kind.mime})
        if filetype.is_image(infile):
            process_image_file(infile, filename, opts, dest_dir=dest_dir, attrs=attrs)
        # TODO: #32 Generate text file previews
    else:
        attrs.update({"extension": filepath.suffix.replace(".", "")})
        attrs.update({"mime_type": guess_type(filepath)[0]})
    file_props = file_template["mapping"]["mappings"]["properties"].keys()
    analysis_props = analysis_template["mapping"]["mappings"]["properties"].keys()
    file_attrs = {}
    analysis_attrs = {}
    # split attrs into 2 sets for indexing
    for key, value in attrs.items():
        if key in file_props:
            file_attrs.update({key: value})
            if key == "analysis_id":
                analysis_attrs.update({key: value})
        elif key in analysis_props:
            analysis_attrs.update({key: value})
        elif key == "analysis":
            for sub_key, sub_value in value.items():
                if sub_key in analysis_props:
                    analysis_attrs.update({sub_key: sub_value})
    file_location = "%s/%s" % (file_attrs["location"], file_attrs["name"])
    file_attrs.update(
        {"file_id": hashlib.md5(file_location.encode("utf-8")).hexdigest()}
    )
    if not file_attrs.get("date_created", False):
        file_attrs.update({"date_created": TODAY})
    if "url" in file_attrs:
        file_attrs.update({"date_accessed": TODAY})
    return file_attrs, analysis_attrs


def index_files(es, files, taxonomy_name, opts):
    """Index files."""
    file_template = index_template(taxonomy_name, opts)
    analysis_template = analysis_index_template(taxonomy_name, opts)
    for infile in files:
        LOGGER.info("Indexing %s", infile)
        p = Path(infile)
        if p.is_dir():
            LOGGER.warn(
                "Argument to --file must be a valid file path. '%s' is a directory",
                infile,
            )
            continue
        file_attrs, analysis_attrs = process_file(
            infile,
            opts,
            file_template=file_template,
            analysis_template=analysis_template,
        )


def update_analysis_attributes(es, analyses, analysis_attrs, analysis_template):
    """Update analysis attributes ready for reindexing."""
    analysis_id = "analysis-%s" % analysis_attrs["analysis_id"]
    action = "index"
    if analysis_id not in analyses:
        res = document_by_id(es, analysis_id, analysis_template["index_name"])
        if res is not None and analysis_id in res:
            analyses.update({analysis_id: res[analysis_id]})
    if analysis_id in analyses:
        action = "update"
        taxon_id = analysis_attrs.get("taxon_id", None)
        assembly_id = analysis_attrs.get("assembly_id", None)
        if (
            taxon_id
            and taxon_id is not None
            and analyses[analysis_id].get("taxon_id", taxon_id) != taxon_id
        ):
            action = "mismatch"
        elif (
            assembly_id
            and assembly_id is not None
            and analyses[analysis_id].get("assembly_id", assembly_id) != assembly_id
        ):
            action = "mismatch"
        if action == "mismatch":
            LOGGER.warning(
                "taxon/assembly ID mismatch for %s, unable to import associated files",
                analysis_id,
            )
            return False
        LOGGER.info("Updating %s", analysis_id)
        # update analysis
        for key, value in analysis_attrs.items():
            if value != "" and value != []:
                if key not in analyses[analysis_id]:
                    analyses[analysis_id].update({key: value})
                elif value != analyses[analysis_id][key]:
                    LOGGER.info(
                        "Updating %s value for %s to '%s'", key, analysis_id, value
                    )
    else:
        analyses[analysis_id] = {}
        for key, value in analysis_attrs.items():
            if value != "":
                analyses[analysis_id].update({key: value})
        # set file counter
        analyses[analysis_id].update({"file_count": 0})
    return action


def update_file_attributes(es, files, file_attrs, analyses, file_template):
    """Update file attributes ready for reindexing."""
    file_id = "file-%s" % file_attrs["file_id"]
    action = "index"
    if file_id not in files:
        res = document_by_id(
            es,
            file_id,
            file_template["index_name"],
        )
        if res is not None and file_id in res:
            files.update({file_id: res[file_id]})
    if file_id in files:
        action = "update"
        analysis_id = file_attrs.get("analysis_id", None)
        if files[file_id].get("analysis_id", analysis_id) != analysis_id:
            LOGGER.warning(
                "analysis ID mismatch for %s, unable to import %s",
                file_id,
                file_attrs.get("name", None),
            )
            return False
        LOGGER.info("Updating %s", file_id)
        # update analysis
        for key, value in file_attrs.items():
            if value != "":
                if key not in files[file_id]:
                    files[file_id].update({key: value})
                elif value != files[file_id][key]:
                    LOGGER.info("Updating %s value for %s to '%s'", key, file_id, value)
    else:
        files[file_id] = {}
        for key, value in file_attrs.items():
            if value != "":
                files[file_id].update({key: value})
        analyses["analysis-%s" % file_attrs["analysis_id"]]["file_count"] += 1
    return action


def stream_docs(doc_dict):
    """Stream a dict of docs."""
    for doc_id, doc in doc_dict.items():
        yield doc_id, doc


def index_docs(es, doc_collection, template, *, dry_run=False):
    """Index a collection of docs."""
    for op_type, doc_dict in doc_collection.items():
        if doc_dict:
            docs = stream_docs(doc_dict)
            index_stream(
                es, template["index_name"], docs, _op_type=op_type, dry_run=dry_run
            )


def check_taxa_exist(es, attrs, taxonomy_name, opts):
    """Check that the specified taxon_id(s) exist in the taxon index."""
    taxon_template = taxon_index_template(taxonomy_name, opts)
    taxon_ids = attrs.get("taxon_id", False)
    if not taxon_ids:
        return {}
    if not isinstance(taxon_ids, list):
        taxon_ids = [taxon_ids]
    taxa = {}
    if taxon_ids:
        taxa = find_or_create_taxa(
            es, opts, taxon_ids=taxon_ids, taxon_template=taxon_template
        )
    if len(taxon_ids) > len(taxa.keys()):
        found = taxa.keys()
        missing = ", ".join(
            [taxon_id for taxon_id in taxon_ids if taxon_id not in found]
        )
        LOGGER.warn("Taxon ID(s) '%s' not found in the specified taxonomy.", missing)
        LOGGER.warn("Unable to import analysis '%s'.", attrs["analysis_id"])
        return None
    return taxa


def check_assemblies_exist(es, attrs, taxonomy_name, opts):
    """Check that the specified taxon_id(s) exist in the taxon index."""
    # similar to check_taxa_exist but for assemblies
    assembly_template = assembly_index_template(taxonomy_name, opts)
    assembly_ids = attrs.get("assembly_id", False)
    if not isinstance(assembly_ids, list):
        assembly_ids = [assembly_ids]
    assemblies = {}
    if assembly_ids:
        assemblies = lookup_assemblies_by_assembly_id(
            es,
            ["assembly-%s" % assembly_id for assembly_id in assembly_ids],
            assembly_template,
            return_type="dict",
        )
    return assemblies


def index_metadata(es, file, taxonomy_name, opts, *, dry_run=False):
    """Index file metadata."""
    data = tofile.load_yaml(file)
    file_template = index_template(taxonomy_name, opts)
    analysis_template = analysis_index_template(taxonomy_name, opts)
    if data is None:
        LOGGER.warn("Unable to load file metadata from '%s'" % file)
        return
    analyses = {}
    files = {}
    analysis_docs = {"index": {}, "update": {}}
    file_docs = {"index": {}, "update": {}}
    for meta in data:
        local = "symlink"
        if "path" in meta:
            if meta["path"].startswith("~"):
                infile = os.path.expanduser(meta["path"])
            else:
                infile = meta["path"]
            del meta["path"]
        elif "url" in meta:
            try:
                infile = tofetch.fetch_tmp_file(meta["url"])
                local = None
            except urllib.error.HTTPError as err:
                LOGGER.warn("Got %s error for %s" % (str(err.code), meta["url"]))
                continue
        elif "name" in meta:
            infile = meta["name"]
        else:
            LOGGER.warn("Found a record with no associated file in '%s'" % file)
        if "name" in meta:
            filename = meta["name"]
            del meta["name"]
        else:
            filename = None
        local = meta.pop("local", local)
        file_attrs, analysis_attrs = process_file(
            infile,
            opts,
            file_template=file_template,
            analysis_template=analysis_template,
            filename=filename,
            meta=meta,
            local=local,
        )
        if "analysis_id" not in analysis_attrs:
            file_name = file_attrs.get("name", file_attrs["file_id"])
            LOGGER.warn("Unable to import %s: analysis_id must be specified", file_name)
            continue
        # TODO: #30 check taxon_id(s) and assembly_id(s) exist in database
        assemblies = check_assemblies_exist(es, analysis_attrs, taxonomy_name, opts)
        if assemblies is None:
            continue
        if "taxon_id" in analysis_attrs and analysis_attrs["taxon_id"]:
            taxa = check_taxa_exist(es, analysis_attrs, taxonomy_name, opts)
        else:
            # set taxa from assembly index
            taxa = {assembly["taxon_id"]: assembly for assembly in assemblies.values()}
            analysis_attrs["taxon_id"] = list(taxa.keys())
        if taxa is None:
            continue
        # TODO: #33 include lineage summary in analysis index
        analysis_attrs["ancestors"] = list_ancestors(taxa)
        # Fetch existing analysis entry if available
        action = update_analysis_attributes(
            es, analyses, analysis_attrs, analysis_template
        )
        if not action:
            # TODO: remove analysis and files with shared analysis name
            continue
        analysis_id = "analysis-%s" % analysis_attrs["analysis_id"]
        if analysis_id in analysis_docs["index"]:
            action = "index"
        analysis_docs[action].update({analysis_id: analyses[analysis_id]})
        # prepare files for indexing (include check for dupes and increment analysis file count)
        file_action = update_file_attributes(
            es, files, file_attrs, analyses, file_template
        )
        if not file_action:
            continue
        file_id = "file-%s" % file_attrs["file_id"]
        if analysis_id in analysis_docs["index"]:
            action = "index"
        file_docs[file_action].update({file_id: files[file_id]})
    # Create/update index entry according to returned action
    LOGGER.info("Indexing analyses")
    index_docs(es, analysis_docs, analysis_template, dry_run=dry_run)
    LOGGER.info("Indexing files")
    index_docs(es, file_docs, file_template, dry_run=dry_run)
